from osbot_utils.utils.Files import file_contents, path_combine

import osbot_prefect


class Version:

    FILE_NAME_VERSION = 'version'

    def path_code_root(self):
        return osbot_prefect.path

    def path_version_file(self):
        return path_combine(self.path_code_root(), self.FILE_NAME_VERSION)

    def value(self):
        version = file_contents(self.path_version_file()) or ""
        return version.strip()

version = Version().value()


from datetime import datetime, timezone, timedelta
from enum import Enum

from osbot_prefect.server.Prefect__States import Prefect__States
from osbot_utils.utils.Dev import pprint

from osbot_utils.base_classes.Type_Safe         import Type_Safe
from osbot_prefect.server.Prefect__Rest_API     import Prefect__Rest_API
from osbot_utils.utils.Objects import dict_to_obj




class Prefect__Cloud_API(Type_Safe):
    prefect_rest_api = Prefect__Rest_API()

    def artifacts__create(self, artifact_data):
        return self.prefect_rest_api.create(target='artifacts', data=artifact_data)

    def flow(self, flow_id):
        return self.prefect_rest_api.read(target='flows', target_id=flow_id)

    def flow__create(self, flow_definition):
        return self.prefect_rest_api.create(target='flows', data=flow_definition)

    def flow__delete(self, flow_id):
        return  self.prefect_rest_api.delete(target='flows', target_id=flow_id)

    def flow_run(self, flow_id):
        return self.prefect_rest_api.read(target='flow_runs', target_id=flow_id)

    def flow_run__create(self, flow_run_definition):
        return self.prefect_rest_api.create(target='flow_runs', data=flow_run_definition)

    def flow_run__input(self, flow_run_id, input_data):
        kwargs = dict(target        = 'flow_runs'       ,
                      target_id     = flow_run_id       ,
                      target_action = 'input'           ,
                      target_data   = input_data        )

        return self.prefect_rest_api.update_action(**kwargs)

    def flow_run__set_state(self, flow_run_id, state):
        kwargs = dict(target        = 'flow_runs'       ,
                      target_id     = flow_run_id       ,
                      target_action = 'set_state'       ,
                      target_data   = { 'state': state })

        return self.prefect_rest_api.update_action(**kwargs)

    def flow_run__set_state_type(self, flow_run_id, state_type):
        return self.flow_run__set_state(flow_run_id, {'type': state_type})

    def flow_run__set_state_type__running(self, flow_run_id):
        return self.flow_run__set_state_type(flow_run_id, Prefect__States.RUNNING)

    def flow_run__set_state_type__completed(self, flow_run_id):
        return self.flow_run__set_state_type(flow_run_id, Prefect__States.COMPLETED)

    def flow_run__delete(self, flow_run_id):
        return self.prefect_rest_api.delete(target='flow_runs', target_id=flow_run_id)

    def flow_run__update(self, flow_run_id, flow_run_definition):
        return self.prefect_rest_api.update(target='flow_runs', target_id=flow_run_id, target_data=flow_run_definition)

    def flows(self, limit=5):
        filter_data = {"sort": "CREATED_DESC",
                       "limit": limit}
        return self.prefect_rest_api.filter(target='flows', filter_data=filter_data).data or []

    def flows_ids(self, limit=5):                                       # todo: see if there is a way to get these IDs directly via a GraphQL query
        flows = self.flows(limit=limit)
        return [flow.id for flow in flows]

    def logs__create(self, log_data):
        return self.prefect_rest_api.create(target='logs', data=log_data)

    def logs__filter(self, limit=5):
        filter_data = {"sort": "TIMESTAMP_DESC",
                       "limit": limit}
        return self.prefect_rest_api.filter(target='logs', filter_data=filter_data)

    def task_run(self, task_run_id):
        return self.prefect_rest_api.read(target='task_runs', target_id=task_run_id)


    def task_run__create(self, task_run_definition):
        return self.prefect_rest_api.create(target='task_runs', data=task_run_definition)

    def task_run__set_state(self, task_run_id, state):
        kwargs = dict(target        = 'task_runs'       ,
                      target_id     = task_run_id       ,
                      target_action = 'set_state'       ,
                      target_data   = { 'state': state })

        return self.prefect_rest_api.update_action(**kwargs)

    def task_run__set_state_type(self, task_run_id, state_type):
        return self.task_run__set_state(task_run_id, {'type': state_type})

    def task_run__set_state_type__running(self, task_run_id):
        return self.task_run__set_state_type(task_run_id, Prefect__States.RUNNING)

    def task_run__set_state_type__running__completed(self, task_run_id):
        return self.task_run__set_state_type(task_run_id, Prefect__States.COMPLETED)

    def to_prefect_timestamp(self, date_time):
        return date_time.isoformat()
        #return date_time.strftime("%Y-%m-%dT%H:%M:%S.%f")[:-3] + "Z"

    def to_prefect_timestamp__now_utc(self):
        return self.to_prefect_timestamp(datetime.now(timezone.utc))

    def to_prefect_timestamp__now_utc__with_delta(self, hours=0, minutes=0, seconds=0, milliseconds=0,microseconds=0):
        current_time = datetime.now(timezone.utc)
        new_time     = current_time + timedelta(hours=hours,minutes=minutes, seconds=seconds,milliseconds=milliseconds, microseconds=microseconds)
        return self.to_prefect_timestamp(new_time)




from types import SimpleNamespace

from osbot_utils.utils.Dev import pprint

from osbot_utils.helpers.Random_Guid import Random_Guid

from osbot_prefect.server.Prefect__Cloud_API import Prefect__Cloud_API
from osbot_utils.utils.Misc import random_text

from osbot_utils.base_classes.Type_Safe import Type_Safe


class Temp__Flow(Type_Safe):
    flow_name         : str             = random_text('pytest-temp-flow')
    flow              : SimpleNamespace = None
    flow_id           : str             = None
    prefect_cloud_api: Prefect__Cloud_API

    def __enter__(self):
        self.flow__create()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.flow__delete()

    def flow__create(self):
        self.flow    = self.prefect_cloud_api.flow__create({'name': self.flow_name}).data
        self.flow_id = self.flow.id
        return self

    def flow__delete(self):
        self.prefect_cloud_api.flow__delete(self.flow_id)

    def flow__exists(self):
        return self.flow__info() is not None

    def flow__info(self):
        return self.prefect_cloud_api.flow(self.flow_id).data




from types import SimpleNamespace

from osbot_prefect.server.Prefect__States import Prefect__States
from osbot_prefect.testing.Temp__Flow import Temp__Flow


class Temp__Flow_Run(Temp__Flow):
    flow_run_id: str             = None
    flow_run   : SimpleNamespace = None

    def __enter__(self):
        super().__enter__()
        self.flow_run__create()
        return self

    def flow_run__create(self):
        self.flow_run    = self.prefect_cloud_api.flow_run__create({'flow_id': self.flow_id}).data
        self.flow_run_id = self.flow_run.id
        return self

    def flow_run__exists(self):
        return self.flow_run__info() is not None

    def flow_run__info(self):
        return self.prefect_cloud_api.flow_run(self.flow_run_id).data

    def flow_run__set_state(self, state):
        return self.prefect_cloud_api.flow_run__set_state_type(self.flow_run_id, state)

    def flow_run__set_state__completed(self):
        return self.flow_run__set_state(Prefect__States.COMPLETED)

    def flow_run__set_state__running(self):
        return self.flow_run__set_state(Prefect__States.RUNNING)


from osbot_utils.base_classes.Type_Safe import Type_Safe

class Prefect__Artifacts(Type_Safe):
    LINKS    : str = "links"
    MARKDOWN : str = "markdown"         # ok: shows in the 'Artifacts' Tab
    PROGRESS : str = "progress"         # todo: check it this one actually works (I was getting 'unknown')
    IMAGES   : str = "images"
    TABLES   : str = "tables"           # todo: also check this one (the prob might be with the structure of the data object)
    RESULT   : str = "result"           # ok: shows in the 'Results' Tab


from osbot_utils.utils.Misc import time_now

from osbot_utils.helpers.Random_Guid import Random_Guid

from osbot_utils.helpers.flows.Task import Task

from osbot_prefect.server.Prefect__States import Prefect__States
from osbot_utils.utils.Dev import pprint

from osbot_prefect.server.Prefect__Cloud_API    import Prefect__Cloud_API
from osbot_utils.helpers.flows.Flow             import Flow
from osbot_utils.base_classes.Type_Safe         import Type_Safe
from osbot_utils.helpers.flows.Flow__Events     import flow_events, Flow__Event_Type, Flow__Event


class Flow_Events__To__Prefect_Server(Type_Safe):
    prefect_cloud_api   : Prefect__Cloud_API
    prefect_ids_mapping : dict

    def __enter__(self):
        self.add_event_listener()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.remove_event_listener()

    def add_event_listener(self):
        flow_events.event_listeners.append(self.event_listener)

    def handle_event(self, event_type: Flow__Event_Type, event_source, event_data):
        if   event_type == Flow__Event_Type.FLOW_START:  self.handle_event__flow_start(flow=event_source)
        elif event_type == Flow__Event_Type.FLOW_STOP :  self.handle_event__flow_stop (flow=event_source)
        elif event_type == Flow__Event_Type.TASK_START:  self.handle_event__task_start(task=event_source)
        elif event_type == Flow__Event_Type.TASK_STOP :  self.handle_event__task_stop (task=event_source)
        else:
            print()
            print(f"Error in handle_event, unknown event_type: {event_type}")

    def handle_event__flow_start(self, flow: Flow):
        prefect__flow_id                         = self.prefect_cloud_api.flow__create({'name': flow.flow_name}).data.id
        tag_1 = 'flow_start'
        tag_2 = time_now()
        prefect__flow_run_definition             = dict(flow_id    = prefect__flow_id                            ,
                                                        name       = flow.flow_id                                ,
                                                        parameters = dict(answer = 42                            ,
                                                                          source = 'handle_event__flow_start'   ),
                                                        context    = dict(context_1 = 42                         ,
                                                                          context_2 = 'handle_event__flow_start'),
                                                        tags       = [tag_1, tag_2                              ])
        prefect_flow_run                         = self.prefect_cloud_api.flow_run__create(prefect__flow_run_definition)
        if prefect_flow_run.status != 'ok':
            pprint("******* Error in handle_event__flow_start ***** ")          # todo: move this to a Flow Events logging system
            pprint(prefect_flow_run)
        else:
            prefect__flow_run_id                     = prefect_flow_run.data.id
            self.prefect_ids_mapping[flow.flow_name] = prefect__flow_id
            self.prefect_ids_mapping[flow.flow_id  ] = prefect__flow_run_id
            self.prefect_cloud_api.flow_run__set_state_type__running(prefect__flow_run_id)

    def handle_event__flow_stop(self, flow: Flow):
        prefect__flow_run_id = self.prefect_ids_mapping.get(flow.flow_id)
        self.prefect_cloud_api.flow_run__set_state_type__completed(prefect__flow_run_id)

    def handle_event__task_start(self, task: Task):
        prefect__flow_run_id          = self.prefect_ids_mapping[task.task_flow.flow_id]
        prefect__task_run_definition  = { 'flow_run_id' : prefect__flow_run_id,
                                          'dynamic_key' : Random_Guid()       ,
                                          'task_key'    : Random_Guid()       ,
                                          'name'        : task.task_name      ,
                                          'task_inputs' : {"prop_1": [{"input_type": "parameter"    ,
                                                                       "name"      : "an-parameter" },
                                                                      {"input_type": "constant"     ,
                                                                        "type"     :"an-type"       }]},
                                          "tags"        : ["tag_a", "tag_b"] }
        prefect__task_run    = self.prefect_cloud_api.task_run__create(prefect__task_run_definition)
        if prefect__task_run.status != 'ok':
            pprint("******* Error in handle_event__task_start ***** ")          # todo: move this to a Flow Events logging system
            pprint(prefect__task_run)
        else:
            prefect__task_run_id = prefect__task_run.data.id
            self.prefect_ids_mapping[task.task_id] = prefect__task_run_id
            self.prefect_cloud_api.task_run__set_state_type__running(prefect__task_run_id)

    def handle_event__task_stop(self, task):
        prefect__task_run_id = self.prefect_ids_mapping.get(task.task_id)
        self.prefect_cloud_api.task_run__set_state_type__running__completed(prefect__task_run_id)

    def event_listener(self, flow_event: Flow__Event):
        event_type   = flow_event.event_type
        event_source = flow_event.event_source
        event_data   = flow_event.event_data
        self.handle_event(event_type=event_type, event_source=event_source, event_data=event_data)

    def remove_event_listener(self):
        flow_events.event_listeners.remove(self.event_listener)


import requests
from osbot_utils.decorators.methods.cache_on_self import cache_on_self

from osbot_utils.utils.Http             import url_join_safe
from osbot_utils.utils.Env              import get_env
from osbot_utils.base_classes.Type_Safe import Type_Safe
from osbot_utils.utils.Objects          import dict_to_obj
from osbot_utils.utils.Status           import status_ok, status_error

ENV_NAME__PREFECT_CLOUD__API_KEY      = 'PREFECT_CLOUD__API_KEY'
ENV_NAME__PREFECT_CLOUD__ACCOUNT_ID   = 'PREFECT_CLOUD__ACCOUNT_ID'
ENV_NAME__PREFECT_CLOUD__WORKSPACE_ID = 'PREFECT_CLOUD__WORKSPACE_ID'
ENV_NAME__PREFECT_TARGET_SERVER       = 'PREFECT_TARGET_SERVER'

DEFAULT_URL__PREFECT_TARGET_SERVER    = "http://localhost:4200/api"

class Prefect__Rest_API(Type_Safe):

    # raw request methods
    def prefect_cloud__api_key(self):
        return get_env(ENV_NAME__PREFECT_CLOUD__API_KEY)

    def prefect_cloud__api_url(self):
        api_key      = self.prefect_cloud__api_key()
        if api_key:
            account_id   = self.prefect_cloud__account_id()
            workspace_id =  self.prefect_cloud__workspace_id()
            if account_id and workspace_id:
                return f"https://api.prefect.cloud/api/accounts/{account_id}/workspaces/{workspace_id}/"


    def prefect_cloud__account_id(self):
        return get_env(ENV_NAME__PREFECT_CLOUD__ACCOUNT_ID)

    def prefect_cloud__workspace_id(self):
        return get_env(ENV_NAME__PREFECT_CLOUD__WORKSPACE_ID)

    @cache_on_self
    def prefect_api_url(self):                                  # todo: add back the support for Prefect API Cloud (to be configured by en
        prefect_cloud_url = self.prefect_cloud__api_url()
        if prefect_cloud_url:
            return prefect_cloud_url
        return DEFAULT_URL__PREFECT_TARGET_SERVER

    def get_headers(self):
        return {"Authorization": f"Bearer {self.prefect_cloud__api_key()}"}                            # Create headers dictionary including authorization token

    def requests__for_method(self, method, path, data=None):
        headers  = self.get_headers()
        endpoint = url_join_safe(self.prefect_api_url(), path)                          # Construct the full endpoint URL by joining the base URL with the path

        if method == requests.delete:                                                   # For DELETE requests, pass data as query parameters
            response = method(endpoint, headers=headers, params=data)
        elif method == requests.get:                                                    # For GET requests, pass data as query parameters
            response = method(endpoint, headers=headers, params=data)
        elif method == requests.head:                                                   # For HEAD requests, no payload or parameters are needed
            response = method(endpoint, headers=headers)
        elif method == requests.post:                                                   # For POST requests, pass data as JSON in the request body
            response = method(endpoint, headers=headers, json=data)
        elif method == requests.patch:                                                  # For PATCH requests, pass data as JSON in the request body
            response = method(endpoint, headers=headers, json=data)
        else:
            return status_error("Unsupported request method")                           # Return an error if the method is not supported

        status_code  = response.status_code                                             # Handle the response and return an appropriate result
        content_type = response.headers.get('Content-Type', '')
        if 200 <= status_code < 300:
            if method == requests.head:                                                 # For HEAD requests, return the headers as the response data
                result = status_ok(data=response.headers)
            elif content_type == 'application/json':                                      # For successful JSON responses, return the JSON data
                json_data  = response.json()
                result = status_ok(data=json_data)
            else:
                result = status_ok(data=response.text)                                      # For other successful requests, return the JSON data
        else:
            result = status_error(message=f"{method.__name__.upper()} request to {path}, failed with status {status_code}", error=response.text) # For failed requests, return an error message with status and response text
        return dict_to_obj(result)


    def requests__delete(self, path, params=None):                                      # Wrapper for executing DELETE requests
        return self.requests__for_method(requests.delete, path, data=params)

    def requests__get(self, path, params=None):                                         # Wrapper for executing GET requests
        return self.requests__for_method(requests.get, path, data=params)

    def requests__post(self, path, data):                                               # Wrapper for executing POST requests
        return self.requests__for_method(requests.post, path, data=data)

    def requests__head(self, path):                                                     # Wrapper for executing HEAD requests
        return self.requests__for_method(requests.head, path)

    def requests__update(self, path, data):                                               # Wrapper for executing PATCH requests
        return self.requests__for_method(requests.patch, path, data=data)

    # request helpers

    def create(self, target, data):
        path = f'/{target}'
        return self.requests__post(path, data)

    def delete(self, target, target_id):
        path = f'/{target}/{target_id}'
        return self.requests__delete(path)

    def read(self, target, target_id):
        path = f'/{target}/{target_id}'
        return self.requests__get(path)

    def filter(self, target, filter_data):          # todo: add support for fetching all items
        path = f'/{target}/filter'
        return self.requests__post(path, filter_data)

    def update(self, target, target_id, target_data):
        path = f'/{target}/{target_id}'
        return self.requests__update(path, target_data)

    def update_action(self, target, target_id, target_action, target_data):
        path = f'/{target}/{target_id}/{target_action}'
        return self.requests__post(path, target_data)


from types                                  import SimpleNamespace
from osbot_prefect.server.Prefect__States   import Prefect__States
from osbot_utils.utils.Misc                 import random_text
from osbot_utils.helpers.Random_Guid        import Random_Guid
from osbot_prefect.testing.Temp__Flow_Run   import Temp__Flow_Run


class Temp__Task_Run(Temp__Flow_Run):
    task_run_definition     : dict
    task_run_dynamic_key    : str              = random_text('task-run-dynamic-key')
    task_run                : SimpleNamespace  = None
    task_run_key            : Random_Guid
    task_run_id             : Random_Guid      = None
    task_run__create_result : SimpleNamespace  = None

    def __enter__(self):
        super().__enter__()
        self.task_run__create()
        return self

    def task_run__create(self, **task_run_definition):
        self.task_run_definition = { 'flow_run_id': self.flow_run_id         ,
                                     'dynamic_key': self.task_run_dynamic_key,
                                     'task_key'   : self.task_run_key        ,
                                     **task_run_definition   }
        self.task_run__create_result = self.prefect_cloud_api.task_run__create(self.task_run_definition)
        self.task_run                = self.task_run__create_result.data
        self.task_run_id             = Random_Guid(self.task_run.id)
        return self

    def task_run__exists(self):
        return self.task_run__info() is not None

    def task_run__info(self):
        return self.prefect_cloud_api.task_run(self.task_run_id).data

    def task_run__set_state(self, state):
        return self.prefect_cloud_api.task_run__set_state_type(self.task_run_id, state)

    def task_run__set_state__completed(self):
        return self.task_run__set_state(Prefect__States.COMPLETED)

    def task_run__set_state__running(self):
        return self.task_run__set_state(Prefect__States.RUNNING)


from osbot_utils.base_classes.Type_Safe import Type_Safe


class Prefect__States(Type_Safe):
    SCHEDULED : str = "SCHEDULED"
    PENDING   : str = "PENDING"
    RUNNING   : str = "RUNNING"
    CANCELLING: str = "CANCELLING"
    CANCELLED : str = "CANCELLED"
    COMPLETED : str = "COMPLETED"
    FAILED    : str = "FAILED"
    CRASHED   : str = "CRASHED"
    PAUSED    : str = "PAUSED"